@@ -73,8 +73,8 @@
 (defvar last-coding-system-used)
 (defun epa-file-insert-file-contents (file &optional visit beg end replace)
   (barf-if-buffer-read-only)
-  (if (or beg end)
-      (error "Can't read the file partially."))
+  (if (and visit (or beg end))
+      (error "Attempt to visit less than an entire file"))
   (setq file (expand-file-name file))
   (let ((local-copy (epa-file-run-real-handler #'file-local-copy (list file)))
 	(context (epg-make-context))
@@ -90,21 +90,25 @@
 	  (if replace
 	      (goto-char (point-min)))
 	  (condition-case error
-	      (setq string (decode-coding-string
-			    (epg-decrypt-file context file nil)
-			    'undecided))
+	      (setq string (epg-decrypt-file context file nil))
 	    (error
 	     (if (setq entry (assoc file epa-file-passphrase-alist))
 		 (setcdr entry nil))
 	     (signal 'file-error
 		     (cons "Opening input file" (cdr error)))))
+	  (if (or beg end)
+	      (setq string (substring string (or beg 0) end)))
+	  (setq string
+		(decode-coding-string string
+				      (or coding-system-for-read 'undecided)))
 	  (if (boundp 'last-coding-system-used)
 	      (set-buffer-file-coding-system last-coding-system-used)
 	    (set-buffer-file-coding-system default-buffer-file-coding-system))
-	  (insert string)
-	  (setq length (length string))
-	  (if replace
-	      (delete-region (point) (point-max))))
+	  (save-excursion
+	    (insert string)
+	    (setq length (length string))
+	    (if replace
+		(delete-region (point) (point-max)))))
       (if (and local-copy
 	       (file-exists-p local-copy))
 	  (delete-file local-copy)))
@@ -116,11 +120,12 @@
   (if append
       (error "Can't append to the file."))
   (setq file (expand-file-name file))
-  (let* ((coding-system (if (boundp 'last-coding-system-used)
-			    (condition-case nil
-				(write-region (point-min) (point-max) "/")
-			      (error last-coding-system-used))
-			  buffer-file-coding-system))
+  (let* ((coding-system (or coding-system-for-write
+			    (if (boundp 'last-coding-system-used)
+				(condition-case nil
+				    (write-region (point-min) (point-max) "/")
+				  (error last-coding-system-used))
+			      buffer-file-coding-system)))
 	 (context (epg-make-context))
 	 (coding-system-for-write 'binary)
 	 string entry)
