@@ -39,7 +39,7 @@
   :group 'epg
   :type 'string)
 
-(defconst epg-version-number "0.0.3")
+(defconst epg-version-number "0.0.4")
 
 (defvar epg-user-id nil
   "GnuPG ID of your default identity.")
@@ -48,6 +48,7 @@
   "An alist mapping from key ID to user ID.")
 
 (defvar epg-read-point nil)
+(defvar epg-process-filter-running nil)
 (defvar epg-pending-status-list nil)
 (defvar epg-key-id nil)
 (defvar epg-context nil)
@@ -759,15 +760,14 @@ This function is for internal use only."
 	   ((eq (epg-signature-status signature) 'no-pubkey)
 	    "No public key for "))
      (epg-signature-key-id signature)
-     " "
      (if user-id
-	 (concat (if (stringp user-id)
+	 (concat " "
+		 (if (stringp user-id)
 		     user-id
-		   (epg-decode-dn user-id))
-		 " ")
+		   (epg-decode-dn user-id)))
        "")
      (if (epg-signature-validity signature)
-	 (format "(trust %s)"  (epg-signature-validity signature))
+	 (format " (trust %s)"  (epg-signature-validity signature))
        ""))))
 
 (defun epg-verify-result-to-string (verify-result)
@@ -829,6 +829,8 @@ This function is for internal use only."
     (with-current-buffer buffer
       (make-local-variable 'epg-read-point)
       (setq epg-read-point (point-min))
+      (make-local-variable 'epg-process-filter-running)
+      (setq epg-process-filter-running nil)
       (make-local-variable 'epg-pending-status-list)
       (setq epg-pending-status-list nil)
       (make-local-variable 'epg-key-id)
@@ -861,21 +863,26 @@ This function is for internal use only."
 	(set-buffer (process-buffer process))
 	(goto-char (point-max))
 	(insert input)
-	(goto-char epg-read-point)
-	(beginning-of-line)
-	(while (looking-at ".*\n")	;the input line finished
-	  (save-excursion
-	    (if (looking-at "\\[GNUPG:] \\([A-Z_]+\\) ?\\(.*\\)")
-		(let* ((status (match-string 1))
-		       (string (match-string 2))
-		       (symbol (intern-soft (concat "epg--status-" status))))
-		  (if (member status epg-pending-status-list)
-		      (setq epg-pending-status-list nil))
-		  (if (and symbol
-			   (fboundp symbol))
-		      (funcall symbol epg-context string)))))
-	  (forward-line))
-	(setq epg-read-point (point)))))
+	(unless epg-process-filter-running
+	  (unwind-protect
+	      (progn
+		(setq epg-process-filter-running t)
+		(goto-char epg-read-point)
+		(beginning-of-line)
+		(while (looking-at ".*\n") ;the input line finished
+		  (if (looking-at "\\[GNUPG:] \\([A-Z_]+\\) ?\\(.*\\)")
+		      (let* ((status (match-string 1))
+			     (string (match-string 2))
+			     (symbol (intern-soft (concat "epg--status-"
+							  status))))
+			(if (member status epg-pending-status-list)
+			    (setq epg-pending-status-list nil))
+			(if (and symbol
+				 (fboundp symbol))
+			    (funcall symbol epg-context string))))
+		  (forward-line)
+		  (setq epg-read-point (point))))
+	    (setq epg-process-filter-running nil))))))
 
 (defun epg-read-output (context)
   "Read the output file CONTEXT and return the content as a string."
@@ -1055,7 +1062,11 @@ This function is for internal use only."
 \\([0-9A-Fa-f][0-9A-Fa-f]\\) \\([^ ]+\\) \\([0-9]+\\)"
 		    string)
       (let ((signature (epg-make-signature 'error)))
-	(epg-context-set-result-for context 'verify (list signature))
+	(epg-context-set-result-for
+	 context
+	 'verify
+	 (cons signature
+	       (epg-context-result-for context 'verify)))
 	(epg-signature-set-key-id
 	 signature
 	 (match-string 1 string))
@@ -1770,9 +1781,13 @@ If you are unsure, use synchronous version of this function
 		     (if sign
 			 (cons "--sign"
 			       (apply #'nconc
-				      (mapcar (lambda (signer)
-						(list "-u" signer))
-					      (epg-context-signers context)))))
+				      (mapcar
+				       (lambda (signer)
+					 (list "-u"
+					       (epg-sub-key-id
+						(car (epg-key-sub-key-list
+						      signer)))))
+				       (epg-context-signers context)))))
 		     (apply #'nconc
 			    (mapcar
 			     (lambda (recipient)
